# pyright: reportUnknownMemberType=false
"""Rich content classes."""

from __future__ import annotations

import base64
import io
from typing import TYPE_CHECKING
import urllib.request
import warnings
import xml.etree.ElementTree as ET

import rico._config
import rico._html


if TYPE_CHECKING:
    from typing import Any, Literal

    _ReprResult = str | bytes | dict[str, str | bytes]


try:
    import altair as alt
    import vl_convert as vlc
except ImportError:
    alt = None

try:
    import matplotlib.pyplot as plt
except ImportError:
    plt = None

try:
    import seaborn.objects as so
except ImportError:
    so = None


class ContentBase:
    container: ET.Element

    def __init__(self, class_: str | None = None):
        """Create an empty <div> container.

        Args:
            class_: The container class attribute.
        """
        attrib = {"class": class_} if class_ is not None else {}
        self.container = ET.Element("div", attrib=attrib)

    def serialize(
        self,
        indent: bool | None = None,
        space: str | None = None,
        strip: bool | None = None,
    ) -> str:
        """Serialize the object's HTML elements to a string.

        Args:
            indent: If True, indent the elements.
            space: Whitespace for indentation.
            strip: If True, strip unnecessary whitespace.

        Returns:
            A string with serialized HTML.
        """
        return rico._html.serialize_html(
            self.container, indent=indent, space=space, strip=strip)

    def __str__(self) -> str:
        """Serialize the object's HTML elements to a string."""
        return self.serialize()


class Tag(ContentBase):
    def __init__(
        self,
        tag: str,
        text: str | None = None,
        tail: str | None = None,
        attrib: dict[str, Any] = {},
        class_: str | None = None,
        **extra: Any,
    ):
        """Create an HTML element using tag parameters and wrap it in a <div> container.

        Args:
            tag: The tag name.
            text: Text before first subelement.
            tail: Text after the closing tag.
            attrib: The tag's attributes.
            extra: Extra attributes.
            class_: The container class attribute.
        """
        super().__init__(class_)
        element = ET.Element(tag, attrib={**attrib, **extra})
        element.text = text
        element.tail = tail
        self.container.append(element)


class Text(ContentBase):
    def __init__(
        self,
        obj: Any,
        mono: bool | None = None,
        wrap: bool | None = None,
        class_: str | None = None,
    ):
        """Create an HTML element from text and wrap it in a <div> container.

        The default tag is <p> if the text doesn't contains a line break.
        Otherwise, the <pre> tag is used.

        The `mono` and `wrap` parameters rely on Bootstrap CSS.

        Args:
            obj: The text. If it's not an instance of str, then it's converted to str.
            mono: If True, set the text font to monospaced.
            wrap: If True, wrap the text.
            class_: The container class attribute.
        """
        super().__init__(class_)

        global_config = rico._config.get_config()
        if mono is None:
            mono = global_config["text_mono"]
        if wrap is None:
            wrap = global_config["text_wrap"]

        if not isinstance(obj, str):
            obj = str(obj)

        text_class = " ".join([
            cl
            for cond, cl in [(mono, "font-monospace"), (wrap, "text-wrap")]
            if cond
        ])

        tag = "pre" if "\n" in obj else "p"
        attrib = {"class": text_class} if text_class else {}
        element = ET.Element(tag, attrib=attrib)
        element.text = obj
        self.container.append(element)


class Code(ContentBase):
    def __init__(self, text: str, class_: str | None = None):
        """Create HTML elements from code and wrap them in a <div> container.

        Args:
            text: The code.
            class_: The container class attribute.
        """
        super().__init__(class_)
        pre = ET.Element("pre")
        code = ET.Element("code")
        code.text = text
        pre.append(code)
        self.container.append(pre)


class HTML(ContentBase):
    def __init__(
        self,
        text: str,
        strip_dataframe_borders: bool = False,
        class_: str | None = None,
    ):
        """Create HTML elements from raw HTML and wrap them in a <div> container.

        Args:
            text: A string with raw HTML.
            strip_dataframe_borders: If True, remove borders attributes from dataframes.
            class_: The container class attribute.
        """
        super().__init__(class_)

        for element in rico._html.parse_html(text):
            self.container.append(element)

        if strip_dataframe_borders:
            for table in self.container.iterfind(
                './/table[@class="dataframe"][@border]',
            ):
                del table.attrib["border"]


class Markdown(ContentBase):
    def __init__(
        self,
        text: str,
        class_: str | None = None,
        **kwargs: Any,
    ):
        """Create HTML elements from Markdown text and wrap them in a <div> container.

        Args:
            text: The Markdown text.
            class_: The container class attribute.
            **kwargs: Keyword arguments passed to the `markdown.markdown` function.

        Raises:
            RuntimeError: The markdown renderer is not defined.
        """
        md_renderer = rico._config.get_config("markdown_renderer")
        if md_renderer is None:
            raise RuntimeError("The markdown renderer is not defined.")

        content = HTML(md_renderer(text, **kwargs), class_=class_)
        self.container = content.container


class Image(ContentBase):
    def __init__(
        self,
        data: bytes | str,
        mime_subtype: str,
        class_: str | None = None,
    ):
        """Create an HTML element from image data and wrap it in a <div> container.

        The image is byte64-encoded.

        Args:
            data: The image data.
            mime_subtype: The image MIME subtype. Examples: "png", "svg+xml".
            class_: The container class attribute.
        """
        super().__init__(class_)

        if isinstance(data, str):
            data = data.encode()
        encoded_image = base64.b64encode(data).decode()

        element = ET.Element(
            "img",
            attrib={"src": f"data:image/{mime_subtype};base64,{encoded_image}"},
        )
        self.container.append(element)


class Plot(ContentBase):
    def __init__(
        self,
        obj: Any,
        format: Literal["svg", "png"] | None = None,  # noqa: A002
        class_: str | None = None,
        **kwargs: Any,
    ):
        """Create an HTML element from a plot object and wrap it in a <div> container.

        The supported plot types are the following:
        - Matplotlib Pyplot Axes and Figure,
        - [Deprecated] Altair Chart,
        - [Deprecated] Seaborn Plot (seaborn.objects interface).

        Deprecated:
            Support of the Seaborn plots in this class/method is deprecated
            and will be removed in version 0.4.0.
            Use the rico.Obj or obj.append indstead.

        Args:
            obj: The plot object.
            format: Image format.
            class_: The container class attribute.
            **kwargs: Keyword arguments passed to the function
                which converts the object to an image.

        Raises:
            TypeError: The plot type is not supported
                or a required extra package is not installed.
        """
        if format is None:
            format = rico._config.get_config("image_format")  # noqa: A001

        if plt is not None and isinstance(obj, plt.Axes):
            obj = obj.figure

        if plt is not None and isinstance(obj, plt.Figure):  # type: ignore
            stream = io.StringIO() if format == "svg" else io.BytesIO()
            obj.savefig(stream, format=format, **kwargs)
            image = stream.getvalue()
        elif so is not None and isinstance(obj, so.Plot):
            warnings.warn(
                "Support of the Seaborn plots in this class/method is deprecated "
                    "and will be removed in version 0.4.0. "
                    "Use the rico.Obj or obj.append indstead.",
                DeprecationWarning,
                stacklevel=2,
            )
            stream = io.StringIO() if format == "svg" else io.BytesIO()
            obj.save(stream, format=format, **kwargs)
            image = stream.getvalue()
        elif alt is not None and isinstance(obj, alt.TopLevelMixin):
            warnings.warn(
                "Support of the Altair plots in this class/method is deprecated "
                    "and will be removed in version 0.4.0. "
                    "Use the rico.Obj or obj.append indstead.",
                DeprecationWarning,
                stacklevel=2,
            )
            convert = vlc.vegalite_to_svg if format == "svg" else vlc.vegalite_to_png  # type: ignore  # noqa: E501
            image = convert(  # type: ignore
                obj.to_json(),  # type: ignore
                vl_version="_".join(alt.SCHEMA_VERSION.split(".")[:2]),
                **kwargs,
            )
        else:
            error_msg = (
                f"The plot type {type(obj)} is not supported "
                "or a required extra package is not installed."
            )
            raise TypeError(error_msg)

        mime_subtype = "svg+xml" if format == "svg" else format
        content = Image(data=image, mime_subtype=mime_subtype, class_=class_)  # type: ignore  # noqa: E501
        self.container = content.container

Chart = Plot


class Script(ContentBase):
    footer: bool = False

    def __init__(
        self,
        text: str | None = None,
        src: str | None = None,
        inline: bool | None = None,
        defer: bool = False,
        attrib: dict[str, Any] = {},
        **extra: Any,
    ):
        """Create an HTML script element.

        Either `text` or `src` should be used.

        Args:
            text: The script text.
            src: The script source link.
            inline: If True, load script inline, downdload it from source link
                and use it as `text`.
            defer: The "defer" attribute of the script. For an inline script, defines
                whether it should be placed in the footer of the document aftert all
                other content.
            attrib: The script attributes.
            **extra: Extra attributes.

        Raises:
            ValueError: Both text and src are not None.
            ValueError: Both text and src are None.
        """
        if inline is None:
            inline = rico._config.get_config("inline_scripts")

        if text is not None and src is not None:
            raise ValueError("Either `text` or `src` should be None.")

        if text is None and src is None:
            raise ValueError("Either `text` or `src` should be not None.")

        if inline and src is not None:
            with urllib.request.urlopen(src) as response:
                text = response.read().decode()
            src = None

        if src is not None:
            attrib = {"src": src, **attrib}
            if defer:
                attrib = {"defer": True, **attrib}
        else:
            self.footer = defer

        self.container = ET.Element("script", {**attrib, **extra})
        self.container.text = text


class Style(ContentBase):
    def __init__(
        self,
        text: str | None = None,
        src: str | None = None,
        inline: bool | None = None,
        attrib: dict[str, Any] = {},
        **extra: Any,
    ):
        """Create an HTML style element.

        Either `text` or `src` should be used.

        Args:
            text: The stylesheet text.
            src: The stylesheet source link.
            inline: If True, load stylesheet inline, downdload it from source link
                and use it as `text`.
            attrib: The stylesheet attributes.
            **extra: Extra attributes.

        Raises:
            ValueError: Both text and src are not None.
            ValueError: Both text and src are None.
        """
        if inline is None:
            inline = rico._config.get_config("inline_scripts")

        if text is not None and src is not None:
            raise ValueError("Either `text` or `src` should be None.")

        if text is None and src is None:
            raise ValueError("Either `text` or `src` should be not None.")

        if inline and src is not None:
            with urllib.request.urlopen(src) as response:
                text = response.read().decode()
            src = None

        if src is not None:
            tag = "link"
            attrib = {"href": src, **attrib}
            if "rel" not in attrib:
                attrib = {**attrib, "rel": "stylesheet"}
        else:
            tag = "style"

        self.container = ET.Element(tag, {**attrib, **extra})
        self.container.text = text


def _call_repr(
    obj: Any,
    repr_name: str,
    mime_type: str = "text/plain",
) -> dict[str, str | bytes]:
    if hasattr(obj, repr_name):
        repr_fn = getattr(obj, repr_name)
        if callable(repr_fn):
            data : _ReprResult | tuple[_ReprResult, _ReprResult] | None = repr_fn()
            if isinstance(data, tuple):
                data = data[0]
            if data is None:
                return {}
            if not isinstance(data, dict):
                return {mime_type: data}
            return data
    return {}

def _decode(data: str | bytes) -> str:
    return data.decode() if isinstance(data, bytes) else data

def _get_repr(obj: Any) -> ET.Element:  # noqa: C901, PLR0911, PLR0912
    data = _call_repr(obj, "_repr_mimebundle_")
    if data == {}:
        data = _call_repr(obj, "_repr_javascript_", "application/javascript")
    if data == {}:
        data = _call_repr(obj, "_repr_html_", "text/html")
    if data == {}:
        data = _call_repr(obj, "_repr_markdown_", "text/markdown")
    if data == {}:
        data = _call_repr(obj, "_repr_svg_", "image/svg+xml")
    if data == {}:
        data = _call_repr(obj, "_repr_png_", "image/png")
    if data == {}:
        data = _call_repr(obj, "_repr_jpeg_", "image/jpeg")

    if "application/javascript" in data and data["application/javascript"]:
        return Script(_decode(data["application/javascript"])).container
    if "text/html" in data and data["text/html"]:
        return HTML(_decode(data["text/html"]), strip_dataframe_borders=True).container
    if "text/markdown" in data and data["text/markdown"]:
        return Markdown(_decode(data["text/markdown"])).container
    if "image/svg+xml" in data and data["image/svg+xml"]:
        return Image(data["image/svg+xml"], "svg+xml").container
    if "image/png" in data and data["image/png"]:
        return Image(data["image/png"], "png").container
    if "image/jpeg" in data and data["image/jpeg"]:
        return Image(data["image/jpeg"], "jpeg").container
    if "image/gif" in data and data["image/gif"]:
        return Image(data["image/gif"], "gif").container
    if "text/plain" in data and data["text/plain"]:
        return Text(data["text/plain"]).container
    return Text(str(obj)).container

class Obj(ContentBase):
    def __init__(self, *objects: Any, class_: str | None = None):
        """Create HTML elements from objects and wrap them in a <div> container.

        Automatically determines the content type in the following order:
        1. `rico` content classes (subclasses of `ContentBase`).
        2. Matplotlib Pyplot.
        3. Dataframes and other types IPython rich replresentation methods.
        All other types are converted to text.

        More about IPython rich representation methods:
        https://ipython.readthedocs.io/en/stable/config/integrating.html

        Args:
            *objects: The objects.
            class_: The container class attribute.
        """
        super().__init__(class_=class_)
        for obj in objects:
            if isinstance(obj, ContentBase):
                container = obj.container
            elif plt is not None and isinstance(obj, plt.Axes | plt.Figure):  # type: ignore  # noqa: E501
                container = Plot(obj).container
            else:
                container = _get_repr(obj)
            self.container.append(container)



# ruff: noqa: F401
# pyright: reportUnusedImport=false
"""Rich content to HTML as easy as Doc(df, plot).

A Python package for creating HTML documents from rich content:
dataframes, plots, images, markdown etc.
It provides a high-level, easy-to-use API with reasonable defaults,
as well as low-level access for better control.
"""

from rico._config import config_context, get_config, set_config
from rico._container import Div, Doc
from rico._content import (
    Chart,
    Code,
    ContentBase,
    HTML,
    Image,
    Markdown,
    Obj,
    Plot,
    Script,
    Style,
    Tag,
    Text,
)
from rico._html import indent_html, parse_html, serialize_html, strip_html
from rico._version import __version__



# pyright: reportPrivateUsage=false
from __future__ import annotations

import importlib
from typing import TYPE_CHECKING
import unittest.mock

import markdown
import markdown_it
import pytest

import rico._config


if TYPE_CHECKING:
    from collections.abc import Callable, Generator
    from typing import Any


def assert_fn_eq(fn1: Callable[..., Any], fn2: Callable[..., Any]) -> None:
    assert fn1.__annotations__ == fn2.__annotations__
    assert fn1.__code__ == fn2.__code__
    assert fn1.__defaults__ == fn2.__defaults__
    assert fn1.__dict__ == fn2.__dict__
    assert fn1.__doc__ == fn2.__doc__
    assert fn1.__kwdefaults__ == fn2.__kwdefaults__
    assert fn1.__module__ == fn2.__module__
    assert fn1.__name__ == fn2.__name__
    assert fn1.__qualname__ == fn2.__qualname__

def test_markdown_import():
    md_renderer = rico._config._global_config["markdown_renderer"]
    assert callable(md_renderer)
    assert_fn_eq(md_renderer, markdown_it.MarkdownIt().render)

    with unittest.mock.patch.dict("sys.modules", {"markdown_it": None}):
        importlib.reload(rico._config)
        md_renderer = rico._config._global_config["markdown_renderer"]
        assert callable(md_renderer)
        assert_fn_eq(md_renderer, markdown.markdown)

    with unittest.mock.patch.dict(
        "sys.modules",
        {"markdown_it": None, "markdown": None},
    ):
        importlib.reload(rico._config)
        md_renderer = rico._config._global_config["markdown_renderer"]
        assert md_renderer is None

    importlib.reload(rico._config)


@pytest.fixture
def reset_config() -> Generator[None, Any, None]:  # noqa: PT004
    old_config = rico._config._global_config.copy()

    try:
        yield
    finally:
        rico._config._global_config = old_config


@pytest.mark.usefixtures("reset_config")
def test_get_config():
    config = rico._config.get_config()
    assert config == rico._config._global_config
    config["indent_html"] = not config["indent_html"]
    assert config != rico._config._global_config

    assert (
        rico._config.get_config("indent_html") ==
        rico._config._global_config["indent_html"]
    )


@pytest.mark.usefixtures("reset_config")
def test_set_config():
    rico._config.set_config(indent_html=True)
    assert rico._config._global_config["indent_html"] is True

    rico._config.set_config(indent_html=False)
    assert rico._config._global_config["indent_html"] is False


@pytest.mark.usefixtures("reset_config")
def test_config_context():
    old_indent_html = rico._config._global_config["indent_html"]

    with rico._config.config_context(indent_html=not old_indent_html):
        assert rico._config._global_config["indent_html"] is not old_indent_html

    assert rico._config._global_config["indent_html"] is old_indent_html



# ruff: noqa: ARG001, PLR0913
"""Global config."""

from __future__ import annotations

import contextlib
import textwrap
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from collections.abc import Callable, Generator
    from typing import Any, Literal


try:
    import markdown_it
    md_renderer = markdown_it.MarkdownIt().render
except ImportError:
    try:
        import markdown
        md_renderer = markdown.markdown
    except ImportError:
        md_renderer = None


BOOTSTRAP_VER = "5"
BOOTSTRAP_CSS= f"https://cdn.jsdelivr.net/npm/bootstrap@{BOOTSTRAP_VER}/dist/css/bootstrap.min.css"
BOOTSTRAP_JS = f"https://cdn.jsdelivr.net/npm/bootstrap@{BOOTSTRAP_VER}/dist/js/bootstrap.min.js"

DATAFRAME_STYLE = textwrap.dedent("""\
    .dataframe table {
        border: none;
        border-collapse: collapse;
        border-spacing: 0;
        margin-bottom: 1em;
        margin-left: 0;
        margin-right: 0;
    }

    .dataframe thead {
        border-bottom: var(--bs-border-width) solid var(--bs-border-color);
        vertical-align: bottom;
    }

    .dataframe td,
    .dataframe th,
    .dataframe tr {
        border: none;
        line-height: var(--bs-body-line-height);
        max-width: none;
        padding: .25rem .5rem;
        text-align: right;
        vertical-align: middle;
        white-space: normal;
    }

    .dataframe th {
        font-weight: bold;
    }
""")


_global_config = {
    "bootstrap_css": BOOTSTRAP_CSS,
    "bootstrap_js": BOOTSTRAP_JS,
    "dataframe_style": DATAFRAME_STYLE,
    "image_format": "svg",
    "indent_html": False,
    "indent_space": "  ",
    "inline_scripts": False,
    "inline_styles": False,
    "markdown_renderer": md_renderer,
    "meta_charset": "utf-8",
    "meta_viewport": "width=device-width, initial-scale=1",
    "strip_html": False,
    "text_mono": False,
    "text_wrap": False,
}


def get_config(option: str | None = None) -> Any:
    """Get global configuration.

    Args:
        option: The option name.

    Returns:
        The value of the option if it's not None,
        or a dictionary with all options otherwise.
    """
    if option is not None:
        return _global_config[option]
    return _global_config.copy()


def set_config(
    bootstrap_css: str | None = None,
    bootstrap_js: str | None = None,
    dataframe_style: str | None = None,
    image_format: Literal["svg", "png"] | None = None,
    indent_html: bool | None = None,
    indent_space: str | None = None,
    inline_scripts: bool | None = None,
    inline_styles: bool | None = None,
    markdown_renderer: Callable[..., str] | None = None,
    meta_charset: str | None = None,
    meta_viewport: str | None = None,
    strip_html: bool | None = None,
    text_mono: bool | None = None,
    text_wrap: bool | None = None,
) -> None:
    """Set global configuration.

    Args:
        bootstrap_css: A link to the Bootstrap CSS file.
            If empty, then the Bootstrap CSS is not loaded.
        bootstrap_js: A link to the Bootstrap JS file.
            If empty, then the Bootstrap JS is not loaded.
        dataframe_style: A dataframe style. If empty, then it's not used.
        image_format: Default plot image format.
        indent_html: Default value of the `indent` parameter for serialization methods.
            If True, indent the elements.
        indent_space: Default value of the `space` parameter for serialization methods.
            Whitespace for indentation.
        inline_scripts: Default value of the `inline` parameter of the Script class.
            If True, load script inline, downdload it from source link.
        inline_styles: Default value of the `inline` parameter of the Style class.
            If True, load stylesheet inline, downdload it from source link.
        markdown_renderer: A callable that converts Markdown to HTML.
            Should accept a Markdown string as the first argument and
            return HTML as a string.
        meta_charset: HTML document charset. If empty, then it's not used.
        meta_viewport: HTML document viewport property. If empty, then it's not used.
        strip_html: Default value of the `strip` parameter for serialization methods.
            If True, strip unnecessary whitespace.
        text_mono: Default value of the `mono` parameter of the Text class.
            If True, set the text font to monospaced.
        text_wrap: Default value of the `wrap` parameter of the Text class.
            If True, wrap the text.
    """
    for param, value in locals().items():
        if value is not None:
            _global_config[param] = value


@contextlib.contextmanager
def config_context(
    bootstrap_css: str | None = None,
    bootstrap_js: str | None = None,
    dataframe_style: str | None = None,
    image_format: Literal["svg", "png"] | None = None,
    indent_html: bool | None = None,
    indent_space: str | None = None,
    inline_scripts: bool | None = None,
    inline_styles: bool | None = None,
    markdown_renderer: Callable[..., str] | None = None,
    meta_charset: str | None = None,
    meta_viewport: str | None = None,
    strip_html: bool | None = None,
    text_mono: bool | None = None,
    text_wrap: bool | None = None,
) -> Generator[None, Any, None]:
    """Context manager for configuration.

    Args:
        bootstrap_css: A link to the Bootstrap CSS file.
            If empty, then the Bootstrap CSS is not loaded.
        bootstrap_js: A link to the Bootstrap JS file.
            If empty, then the Bootstrap JS is not loaded.
        dataframe_style: A dataframe style. If empty, then it's not used.
        image_format: Default plot image format.
        indent_html: Default value of the `indent` parameter for serialization methods.
            If True, indent the elements.
        indent_space: Default value of the `space` parameter for serialization methods.
            Whitespace for indentation.
        inline_scripts: Default value of the `inline` parameter of the Script class.
            If True, load script inline, downdload it from source link.
        inline_styles: Default value of the `inline` parameter of the Style class.
            If True, load stylesheet inline, downdload it from source link.
        markdown_renderer: A callable that converts Markdown to HTML.
            Should accept a Markdown string as the first argument and
            return HTML as a string.
        meta_charset: HTML document charset. If empty, then it's not used.
        meta_viewport: HTML document viewport property. If empty, then it's not used.
        strip_html: Default value of the `strip` parameter for serialization methods.
            If True, strip unnecessary whitespace.
        text_mono: Default value of the `mono` parameter of the Text class.
            If True, set the text font to monospaced.
        text_wrap: Default value of the `wrap` parameter of the Text class.
            If True, wrap the text.
    """
    new_config = locals()
    old_config = get_config()
    set_config(**new_config)

    try:
        yield
    finally:
        set_config(**old_config)



"""HTML parser and serializer."""

from __future__ import annotations

import html.parser
import textwrap
import xml.etree.ElementTree as ET

import rico._config


TAGS_EMPTY = {
    "area", "base", "basefont", "br", "col", "embed", "frame", "hr", "img",
    "input", "isindex", "link", "meta", "param", "path", "source", "track", "wbr",
}

TAGS_INLINE = {
    "a", "abbr", "b", "bdi", "bdo", "br", "cite", "code", "data", "dfn", "em",
    "i", "kbd", "mark", "q", "rp", "rt", "ruby", "s", "samp", "small", "span",
    "strong", "sub", "sup", "time", "u", "var", "wbr",
}

TAGS_NOT_ESCAPED = {"script", "style"}
TAGS_PREFORMATTED = {"pre"}


class _HTMLParser(html.parser.HTMLParser):
    def __init__(self):
        super().__init__()
        self._root = "root"
        self._builder = ET.TreeBuilder()
        self._builder.start(self._root, {})
        self._empty_tag = None

    def handle_starttag(
        self,
        tag: str,
        attrs: list[tuple[str, str | None]],
    ) -> None:
        if self._empty_tag:
            self._builder.end(self._empty_tag)
        self._empty_tag = tag if tag.lower() in TAGS_EMPTY else None
        self._builder.start(tag, dict(attrs))

    def handle_endtag(self, tag: str) -> None:
        if self._empty_tag:
            if self._empty_tag.lower() != tag.lower():
                self._builder.end(self._empty_tag)
            self._empty_tag = None
        self._builder.end(tag)

    def handle_data(self, data: str) -> None:
        if self._empty_tag:
            self._builder.end(self._empty_tag)
            self._empty_tag = None
        self._builder.data(data)

    def close(self) -> tuple[ET.Element]:  # type: ignore
        super().close()
        self._builder.end(self._root)
        return tuple(self._builder.close())


def parse_html(data: str) -> tuple[ET.Element]:
    """Parse HTML from a string.

    Assigns None values to boolean attributes.

    Ignores comments, doctype declaration and processing instructions.
    Converts attribute names to lower case, even for SVG.

    Args:
        data: The string containing HTML data.

    Returns:
        HTML elements parsed from the string.
    """
    parser = _HTMLParser()
    parser.feed(data)
    return parser.close()


def indent_html(
    element: ET.Element,
    space: str = "  ",
    level: int = 0,
) -> ET.Element:
    """Indent an HTML element and its chidren.

    Tnserts newlines and indentation space after elements.
    Creates a new element instead of updating in place.
    Doesn't indent elements inside the <pre> tag or inside inline tags.

    Args:
        element: The element to indent.
        space: Whitespace to insert for each indentation level.
        level: The initial indentation level. Should always be 0.

    Returns:
        Indented HTML element.
    """
    indented_element = ET.Element(element.tag, attrib=element.attrib)
    indented_element.text = element.text
    indented_element.tail = element.tail

    if element.tag.lower() in TAGS_PREFORMATTED | TAGS_INLINE:
        for child in element:
            indented_element.append(child)
    else:
        has_children = len(element) > 0
        left_stripped_text = (
            indented_element.text.lstrip()
            if indented_element.text else
            indented_element.text
        )
        if (
            not has_children and
            indented_element.text is not None and
            left_stripped_text
        ):
            if "\n" in left_stripped_text:
                indented_element.text = "\n" + textwrap.indent(
                    textwrap.dedent(indented_element.text).strip(),
                    space * (level + 1),
                ) + "\n" + space * level
            else:
                indented_element.text = left_stripped_text
        elif not left_stripped_text:
            indented_element.text = "\n" + space * (level + 1) if has_children else None

        for child in element:
            indented_child = indent_html(child, space=space, level=level + 1)
            indented_element.append(indented_child)
            if not indented_child.tail or not indented_child.tail.strip():
                indented_child.tail = "\n" + space * (level + 1)

        if has_children:
            last_child = indented_element[-1]
            if not last_child.tail or not last_child.tail.strip():
                last_child.tail = "\n" + space * level

    return indented_element


def strip_html(element: ET.Element) -> ET.Element:
    """Remove unnecessary whitespace from an HTML element and its children.

    Removes leading whitespace from elements' text.
    Doesn't strip elements inside the <pre> tag or inside inline tags.

    Args:
        element: The element to strip.

    Returns:
        An HTML element with stripped leading and trailing whitespace.
    """
    stripped_element = ET.Element(element.tag, attrib=element.attrib)
    stripped_element.text = element.text
    stripped_element.tail = element.tail

    if element.tag.lower() in TAGS_INLINE | TAGS_PREFORMATTED:
        for child in element:
            stripped_element.append(child)
    else:
        stripped_element.text = (
            stripped_element.text.lstrip()
            if stripped_element.text else
            stripped_element.text
        )

        has_children = len(element) > 0
        if stripped_element.text and not has_children:
            stripped_element.text = stripped_element.text.rstrip()

        for child in element:
            stripped_element.append(strip_html(child))

    return stripped_element


def _escape_cdata(text: str) -> str:
    """Copy of xml.etree.ElementTree._escape_cdata."""
    if "&" in text:
        text = text.replace("&", "&amp;")
    if "<" in text:
        text = text.replace("<", "&lt;")
    if ">" in text:
        text = text.replace(">", "&gt;")
    return text

def _escape_attrib_html(text: str) -> str:
    """Copy of xml.etree.ElementTree._escape_attrib_html."""
    if "&" in text:
        text = text.replace("&", "&amp;")
    if ">" in text:
        text = text.replace(">", "&gt;")
    if '"' in text:
        text = text.replace('"', "&quot;")
    return text

def serialize_html(
    element: ET.Element,
    indent: bool | None = None,
    space: str | None = None,
    strip: bool | None = None,
) -> str:
    """Serialize an HTML element and its children to a string.

    Serializes attributes with None values as boolean.

    Args:
        element: The HTML element.
        indent: If True, indent the element.
        space: Whitespace for indentation.
        strip: If True, strip unnecessary whitespace.

    Returns:
        A string with serialized HTML.
    """
    global_config = rico._config.get_config()
    if indent is None:
        indent = global_config["indent_html"]
    if space is None:
        space = global_config["indent_space"]
    if strip is None:
        strip = global_config["strip_html"]

    if strip:
        element = strip_html(element)
    if indent:
        element = indent_html(element, space=space)  # type: ignore

    attrib = "".join(
        f' {k}="{_escape_attrib_html(v)}"'
        if v is not None and not isinstance(v, bool)  # type: ignore
        # Serialize attributes with None values as boolean.
        else f" {k}"
        for k, v in element.items()
        if not isinstance(v, bool) or v
    )

    ltag = element.tag.lower()
    closing_slash = "/" if ltag in TAGS_EMPTY else ""
    opening_tag = f"<{element.tag}{attrib}{closing_slash}>"

    if element.text is not None:
        text = element.text if ltag in TAGS_NOT_ESCAPED else _escape_cdata(element.text)
    else:
        text = ""

    children = "".join(serialize_html(e) for e in element)
    closing_tag = f"</{element.tag}>" if ltag not in TAGS_EMPTY else ""
    tail = _escape_cdata(element.tail) if element.tail is not None else ""

    return opening_tag + text + children + closing_tag + tail



# pyright: reportUnknownMemberType=false, reportUnknownArgumentType=false
from __future__ import annotations

import base64
import importlib
import io
import textwrap
from typing import TYPE_CHECKING
import unittest.mock
import xml.etree.ElementTree as ET

import altair as alt
import matplotlib.pyplot as plt
import pytest
import seaborn.objects as so

import rico._config
import rico._content


if TYPE_CHECKING:
    from typing import Any, Literal


def test_import_error():
    with unittest.mock.patch.dict("sys.modules", {"altair": None}):
        importlib.reload(rico._content)
        assert rico._content.alt is None

    with unittest.mock.patch.dict("sys.modules", {"vl_convert": None}):
        importlib.reload(rico._content)
        assert rico._content.alt is None

    with unittest.mock.patch.dict("sys.modules", {"matplotlib.pyplot": None}):
        importlib.reload(rico._content)
        assert rico._content.plt is None

    with unittest.mock.patch.dict("sys.modules", {"seaborn.objects": None}):
        importlib.reload(rico._content)
        assert rico._content.so is None

    importlib.reload(rico._content)


def test_content_base_simple():
    content = rico._content.ContentBase()
    div = content.container
    assert isinstance(div, ET.Element)
    assert div.tag == "div"
    assert div.attrib == {}
    assert div.text is None
    assert div.tail is None
    assert len(div) == 0


def test_content_base_with_class():
    content = rico._content.ContentBase("row")
    div = content.container
    assert isinstance(div, ET.Element)
    assert div.tag == "div"
    assert div.attrib == {"class": "row"}
    assert div.text is None
    assert div.tail is None
    assert len(div) == 0


@pytest.fixture
def content_base_subclass_sample():
    class ContentBaseSubclass(rico._content.ContentBase):
        def __init__(self, class_: str | None = None):
            super().__init__(class_)
            p = ET.Element("p")
            p.text = "Hello world"
            self.container.append(p)

    return ContentBaseSubclass("row")


def test_content_base_str(content_base_subclass_sample: rico._content.ContentBase):
    expectation = '<div class="row"><p>Hello world</p></div>'
    assert str(content_base_subclass_sample) == expectation


def test_content_base_indent(content_base_subclass_sample: rico._content.ContentBase):
    expectation = textwrap.dedent("""\
        <div class="row">
            <p>Hello world</p>
        </div>""")
    assert content_base_subclass_sample.serialize(
        indent=True, space="    ") == expectation


def test_content_base_strip(content_base_subclass_sample: rico._content.ContentBase):
    expectation = '<div class="row"><p>Hello world</p></div>'
    assert content_base_subclass_sample.serialize(strip=True) == expectation


def test_tag():
    content = rico._content.Tag(
        "p",
        attrib={"class": "col"},
        id="42",
        text="Hello",
        tail="world",
        class_="row",
    )

    div = content.container
    assert isinstance(div, ET.Element)
    assert div.tag == "div"
    assert div.attrib == {"class": "row"}
    assert div.text is None
    assert div.tail is None
    assert len(div) == 1

    p = tuple(div)[0]
    assert isinstance(p, ET.Element)
    assert p.tag == "p"
    assert p.attrib == {"class": "col", "id": "42"}
    assert p.text == "Hello"
    assert p.tail == "world"
    assert len(p) == 0


def test_text_simple():
    content = rico._content.Text("Hello world", class_="row")

    div = content.container
    assert isinstance(div, ET.Element)
    assert div.tag == "div"
    assert div.attrib == {"class": "row"}
    assert div.text is None
    assert div.tail is None
    assert len(div) == 1

    p = tuple(div)[0]
    assert isinstance(p, ET.Element)
    assert p.tag == "p"
    assert p.attrib == {}
    assert p.text == "Hello world"
    assert p.tail is None
    assert len(p) == 0


def test_text_pre_mono():
    content = rico._content.Text("Hello\nworld", mono=True)
    div = content.container

    pre = tuple(div)[0]
    assert isinstance(pre, ET.Element)
    assert pre.tag == "pre"
    assert pre.attrib == {"class": "font-monospace"}
    assert pre.text == "Hello\nworld"
    assert pre.tail is None
    assert len(pre) == 0


def test_text_int_mono_wrap():
    content = rico._content.Text(42, mono=True, wrap=True)
    div = content.container

    p = tuple(div)[0]
    assert isinstance(p, ET.Element)
    assert p.tag == "p"
    assert p.attrib == {"class": "font-monospace text-wrap"}
    assert p.text == "42"
    assert p.tail is None
    assert len(p) == 0


def test_code():
    content = rico._content.Code("Hello world", class_="row")

    div = content.container
    assert isinstance(div, ET.Element)
    assert div.tag == "div"
    assert div.attrib == {"class": "row"}
    assert div.text is None
    assert div.tail is None
    assert len(div) == 1

    pre = tuple(div)[0]
    assert isinstance(pre, ET.Element)
    assert pre.tag == "pre"
    assert pre.attrib == {}
    assert pre.text is None
    assert pre.tail is None
    assert len(pre) == 1

    code = tuple(pre)[0]
    assert isinstance(code, ET.Element)
    assert code.tag == "code"
    assert code.attrib == {}
    assert code.text == "Hello world"
    assert code.tail is None
    assert len(code) == 0


def test_html_simple():
    content = rico._content.HTML('<p border="1">Hello world</p>', True, class_="row")

    div = content.container
    assert isinstance(div, ET.Element)
    assert div.tag == "div"
    assert div.attrib == {"class": "row"}
    assert div.text is None
    assert div.tail is None
    assert len(div) == 1

    p = tuple(div)[0]
    assert isinstance(p, ET.Element)
    assert p.tag == "p"
    assert p.attrib == {"border": "1"}
    assert p.text == "Hello world"
    assert p.tail is None
    assert len(p) == 0


@pytest.mark.parametrize("border", [True, False], ids=["border", "no border"])
@pytest.mark.parametrize("dataframe", [True, False], ids=["dataframe", "not dataframe"])
@pytest.mark.parametrize(
    "strip_dataframe_borders", [True, False], ids=["strip", "do not strip"])
@pytest.mark.parametrize(
    "wrap_in_div", [True, False], ids=["wrap in div", "do not wrap in div"])
def test_html_table_border(
    border: bool,
    dataframe: bool,
    strip_dataframe_borders: bool,
    wrap_in_div: bool,
):
    border_attr = ' border="1"' if border else ""
    dataframe_attr = ' class="dataframe"' if dataframe else ""

    df = (
        f"<table{border_attr}{dataframe_attr}>"
        "<tbody><tr><td>1<td></tr></tbody></table>"
    )

    if wrap_in_div:
        df = f"<div>{df}</div>"

    content = rico._content.HTML(df, strip_dataframe_borders)
    if wrap_in_div:
        div = tuple(content.container)[0]
        table = tuple(div)[0]
    else:
        table = tuple(content.container)[0]

    if border and (not dataframe or not strip_dataframe_borders):
        assert table.get("border") == "1"
    else:
        assert table.get("border", "no border") == "no border"


def test_markdown():
    content = rico._content.Markdown(
        textwrap.dedent("""\
            # Header 1
            ## Header 2
            Hello world"""),
        class_="row",
    )

    div = content.container
    assert isinstance(div, ET.Element)
    assert div.tag == "div"
    assert div.attrib == {"class": "row"}
    assert div.text is None
    assert div.tail is None
    assert len(div) == 3

    h1 = tuple(div)[0]
    assert isinstance(h1, ET.Element)
    assert h1.tag == "h1"
    assert h1.attrib == {}
    assert h1.text == "Header 1"
    assert h1.tail == "\n"
    assert len(h1) == 0

    h2 = tuple(div)[1]
    assert isinstance(h2, ET.Element)
    assert h2.tag == "h2"
    assert h2.attrib == {}
    assert h2.text == "Header 2"
    assert h2.tail == "\n"
    assert len(h2) == 0

    p = tuple(div)[2]
    assert isinstance(p, ET.Element)
    assert p.tag == "p"
    assert p.attrib == {}
    assert p.text == "Hello world"
    assert len(p) == 0


def test_markdown_error():
    with unittest.mock.patch.dict(
        "sys.modules",
        {"markdown_it": None, "markdown": None},
    ):
        importlib.reload(rico._config)
        with pytest.raises(RuntimeError):
            rico._content.Markdown("Hello world")

    importlib.reload(rico._config)


svg_data = (
    '<svg xmlns="http://www.w3.org/2000/svg" '
    'xmlns:xlink="http://www.w3.org/1999/xlink" '
    'width="16" height="16" fill="currentColor" class="bi bi-dash">'
    '<path d="M4 8a.5.5 0 0 1 .5-.5h7a.5.5 0 0 1 0 1h-7A.5.5 0 0 1 4 8z"/>'
    "</svg>"
)

@pytest.mark.parametrize("data", [svg_data, svg_data.encode()], ids=["str", "bytes"])
def test_image_svg(data: str | bytes):
    content = rico._content.Image(data, mime_subtype="svg+xml", class_="row")

    if isinstance(data, str):
        data = data.encode()
    encoded_image = base64.b64encode(data).decode()

    div = content.container
    assert isinstance(div, ET.Element)
    assert div.tag == "div"
    assert div.attrib == {"class": "row"}
    assert div.text is None
    assert div.tail is None
    assert len(div) == 1

    img = tuple(div)[0]
    assert isinstance(img, ET.Element)
    assert img.tag == "img"
    assert img.attrib == {"src": f"data:image/svg+xml;base64,{encoded_image}"}
    assert img.text is None
    assert img.tail is None
    assert len(img) == 0


@pytest.mark.parametrize("data", [svg_data, svg_data.encode()], ids=["str", "bytes"])
def test_image_png(data: str | bytes):
    content = rico._content.Image(data, mime_subtype="png")

    if isinstance(data, str):
        data = data.encode()
    encoded_image = base64.b64encode(data).decode()

    div = content.container
    assert isinstance(div, ET.Element)
    assert div.tag == "div"
    assert div.attrib == {}
    assert div.text is None
    assert div.tail is None
    assert len(div) == 1

    img = tuple(div)[0]
    assert isinstance(img, ET.Element)
    assert img.tag == "img"
    assert img.attrib == {"src": f"data:image/png;base64,{encoded_image}"}
    assert img.text is None
    assert img.tail is None
    assert len(img) == 0


altair_chart = alt.Chart(
    alt.Data(values=[
        {"x": "A", "y": 5},
        {"x": "B", "y": 3},
        {"x": "C", "y": 6},
        {"x": "D", "y": 7},
        {"x": "E", "y": 2},
    ]),
).mark_bar().encode(x="x:N", y="y:Q")

pyplot_figure, pyplot_axes = plt.subplots()  # type: ignore
pyplot_axes.plot([1, 2, 3, 4], [1, 4, 2, 3])  # type: ignore

seaborn_plot = so.Plot({"x": [1, 2, 3, 4], "y": [1, 4, 2, 3]})  # type: ignore

@pytest.mark.parametrize(
    "plot",
    [altair_chart, pyplot_axes, pyplot_figure, seaborn_plot],
    ids=["altair", "pyplot_axes", "pyplot_figure", "seaborn_plot"],
)
@pytest.mark.parametrize("format", [None, "png"], ids=["svg", "png"])
def test_plot_complete(plot: Any, format: Literal["svg", "png"] | None):  # noqa: A002
    content = rico._content.Plot(plot, format=format, class_="row")

    div = content.container
    assert isinstance(div, ET.Element)
    assert div.tag == "div"
    assert div.attrib == {"class": "row"}
    assert div.text is None
    assert div.tail is None
    assert len(div) == 1

    img = tuple(div)[0]
    assert isinstance(img, ET.Element)
    assert img.tag == "img"


@pytest.mark.parametrize(
    ("module", "err_plot", "plot"),
    [
        ("alt", altair_chart, seaborn_plot),
        ("plt", pyplot_axes, altair_chart),
        ("so", seaborn_plot, pyplot_axes),
    ],
    ids=["alt", "plt", "so"],
)
def test_plot_error(module: str, err_plot: Any, plot: Any):
    with unittest.mock.patch.object(rico._content, module, None):
        with pytest.raises(TypeError):
            rico._content.Plot(err_plot)

        content = rico._content.Plot(plot, class_="row")
        div = content.container
        assert isinstance(div, ET.Element)


@pytest.mark.parametrize("defer", [True, False], ids=["defer", "not defer"])
def test_script_text(defer: bool):
    text = "alert('Hello World!');"
    attrib = {"async": True}
    content = rico._content.Script(text=text, defer=defer, attrib=attrib)

    script = content.container
    assert isinstance(script, ET.Element)
    assert script.tag == "script"
    assert script.attrib == attrib
    assert script.text == text
    assert script.tail is None
    assert len(script) == 0

    assert content.container == script
    assert content.footer == defer


@pytest.mark.parametrize("defer", [True, False], ids=["defer", "not defer"])
def test_script_src(defer: bool):
    src = "javascript.js"
    attrib = {"async": True}
    content = rico._content.Script(src=src, defer=defer, attrib=attrib)

    attrib = {"src": src, **attrib}
    if defer:
        attrib = {"defer": True, **attrib}

    script = content.container
    assert isinstance(script, ET.Element)
    assert script.tag == "script"
    assert script.attrib == attrib
    assert script.text is None
    assert script.tail is None
    assert len(script) == 0

    assert content.container == script
    assert content.footer is False


@pytest.mark.parametrize("defer", [True, False], ids=["defer", "not defer"])
def test_script_inline(defer: bool):
    text = "alert('Hello World!');"
    src = "javascript.js"
    attrib = {"async": True}

    with unittest.mock.patch("rico._content.urllib.request.urlopen") as urlopen:
        urlopen.return_value = io.BytesIO(text.encode())
        content = rico._content.Script(src=src, inline=True, defer=defer, attrib=attrib)
        urlopen.assert_called_once_with(src)

    script = content.container
    assert isinstance(script, ET.Element)
    assert script.tag == "script"
    assert script.attrib == attrib
    assert script.text == text
    assert script.tail is None
    assert len(script) == 0

    assert content.container == script
    assert content.footer == defer


def test_script_raises():
    with pytest.raises(ValueError):
        rico._content.Script()
    with pytest.raises(ValueError):
        rico._content.Script(src="javascript.js", text="alert('Hello World!');")


def test_style_text():
    text = "p {color: red;}"
    attrib = {"title": "Style title"}
    content = rico._content.Style(text=text, attrib=attrib)

    style = content.container
    assert isinstance(style, ET.Element)
    assert style.tag == "style"
    assert style.attrib == attrib
    assert style.text == text
    assert style.tail is None
    assert len(style) == 0

    assert content.container == style


def test_style_src():
    src = "style.css"
    attrib = {"crossorigin": "anonymous"}
    content = rico._content.Style(src=src, attrib=attrib)

    attrib = {"href": src, **attrib, "rel": "stylesheet"}

    link = content.container
    assert isinstance(link, ET.Element)
    assert link.tag == "link"
    assert link.attrib == attrib
    assert link.text is None
    assert link.tail is None
    assert len(link) == 0

    assert content.container == link


def test_style_inline():
    text = "p {color: red;}"
    src = "style.css"
    attrib = {"title": "Style title"}

    with unittest.mock.patch("rico._content.urllib.request.urlopen") as urlopen:
        urlopen.return_value = io.BytesIO(text.encode())
        content = rico._content.Style(src=src, inline=True, attrib=attrib)
        urlopen.assert_called_once_with(src)

    style = content.container
    assert isinstance(style, ET.Element)
    assert style.tag == "style"
    assert style.attrib == attrib
    assert style.text == text
    assert style.tail is None
    assert len(style) == 0

    assert content.container == style


def test_style_raises():
    with pytest.raises(ValueError):
        rico._content.Style()
    with pytest.raises(ValueError):
        rico._content.Style(src="style.css", text="p {color: red;}")


def test_obj_content_base():
    content = rico._content.Obj(
        rico._content.ContentBase(class_="col"),
        class_="row",
    )

    div0 = content.container
    assert isinstance(div0, ET.Element)
    assert div0.tag == "div"
    assert div0.attrib == {"class": "row"}
    assert div0.text is None
    assert div0.tail is None
    assert len(div0) == 1

    div1 = tuple(div0)[0]
    assert isinstance(div1, ET.Element)
    assert div1.tag == "div"
    assert div1.attrib == {"class": "col"}
    assert div1.text is None
    assert div1.tail is None
    assert len(div1) == 0


def test_obj_pyplot():
    content = rico._content.Obj(pyplot_axes, class_="row")

    div0 = content.container
    assert isinstance(div0, ET.Element)
    assert div0.tag == "div"
    assert div0.attrib == {"class": "row"}
    assert div0.text is None
    assert div0.tail is None
    assert len(div0) == 1

    div1 = tuple(div0)[0]
    assert isinstance(div1, ET.Element)
    assert div1.tag == "div"
    assert div1.attrib == {}
    assert div1.text is None
    assert div1.tail is None
    assert len(div1) == 1

    img = tuple(div1)[0]
    assert isinstance(img, ET.Element)
    assert img.tag == "img"
    assert len(img) == 0


def test_obj_javascript():
    class Repr:
        def _repr_javascript_(self) -> str:
            return "alert('Hello World!');"

    content = rico._content.Obj(Repr(), class_="row")

    div0 = content.container
    assert isinstance(div0, ET.Element)
    assert div0.tag == "div"
    assert div0.attrib == {"class": "row"}
    assert div0.text is None
    assert div0.tail is None
    assert len(div0) == 1

    script = tuple(div0)[0]
    assert isinstance(script, ET.Element)
    assert script.tag == "script"
    assert script.attrib == {}
    assert script.text == "alert('Hello World!');"
    assert script.tail is None
    assert len(script) == 0


def test_obj_html():
    class Repr:
        def _repr_html_(self) -> str:
            return "<p>Hello world!</p>"

    content = rico._content.Obj(Repr(), class_="row")

    div0 = content.container
    assert isinstance(div0, ET.Element)
    assert div0.tag == "div"
    assert div0.attrib == {"class": "row"}
    assert div0.text is None
    assert div0.tail is None
    assert len(div0) == 1

    div1 = tuple(div0)[0]
    assert isinstance(div1, ET.Element)
    assert div1.tag == "div"
    assert div1.attrib == {}
    assert div1.text is None
    assert div1.tail is None
    assert len(div1) == 1

    p = tuple(div1)[0]
    assert isinstance(p, ET.Element)
    assert p.tag == "p"
    assert p.attrib == {}
    assert p.text == "Hello world!"
    assert p.tail is None
    assert len(p) == 0


def test_obj_markdown():
    class Repr:
        def _repr_markdown_(self) -> str:
            return "# Hello world!"

    content = rico._content.Obj(Repr(), class_="row")

    div0 = content.container
    assert isinstance(div0, ET.Element)
    assert div0.tag == "div"
    assert div0.attrib == {"class": "row"}
    assert div0.text is None
    assert div0.tail is None
    assert len(div0) == 1

    div1 = tuple(div0)[0]
    assert isinstance(div1, ET.Element)
    assert div1.tag == "div"
    assert div1.attrib == {}
    assert div1.text is None
    assert div1.tail is None
    assert len(div1) == 1

    h1 = tuple(div1)[0]
    assert isinstance(h1, ET.Element)
    assert h1.tag == "h1"
    assert h1.attrib == {}
    assert h1.text == "Hello world!"
    assert len(h1) == 0


def test_obj_svg():
    image = svg_data
    class Repr:
        def _repr_svg_(self) -> str:
            return image

    content = rico._content.Obj(Repr(), class_="row")

    div0 = content.container
    assert isinstance(div0, ET.Element)
    assert div0.tag == "div"
    assert div0.attrib == {"class": "row"}
    assert div0.text is None
    assert div0.tail is None
    assert len(div0) == 1

    div1 = tuple(div0)[0]
    assert isinstance(div1, ET.Element)
    assert div1.tag == "div"
    assert div1.attrib == {}
    assert div1.text is None
    assert div1.tail is None
    assert len(div1) == 1

    encoded_image = base64.b64encode(image.encode()).decode()

    img = tuple(div1)[0]
    assert isinstance(img, ET.Element)
    assert img.tag == "img"
    assert img.attrib == {"src": f"data:image/svg+xml;base64,{encoded_image}"}
    assert img.text is None
    assert img.tail is None
    assert len(img) == 0


def test_obj_png():
    image = svg_data.encode()
    class Repr:
        def _repr_png_(self) -> bytes:
            return image

    content = rico._content.Obj(Repr(), class_="row")

    div0 = content.container
    assert isinstance(div0, ET.Element)
    assert div0.tag == "div"
    assert div0.attrib == {"class": "row"}
    assert div0.text is None
    assert div0.tail is None
    assert len(div0) == 1

    div1 = tuple(div0)[0]
    assert isinstance(div1, ET.Element)
    assert div1.tag == "div"
    assert div1.attrib == {}
    assert div1.text is None
    assert div1.tail is None
    assert len(div1) == 1

    encoded_image = base64.b64encode(image).decode()

    img = tuple(div1)[0]
    assert isinstance(img, ET.Element)
    assert img.tag == "img"
    assert img.attrib == {"src": f"data:image/png;base64,{encoded_image}"}
    assert img.text is None
    assert img.tail is None
    assert len(img) == 0


def test_obj_jpeg():
    image = svg_data.encode()
    class Repr:
        def _repr_jpeg_(self) -> bytes:
            return image

    content = rico._content.Obj(Repr(), class_="row")

    div0 = content.container
    assert isinstance(div0, ET.Element)
    assert div0.tag == "div"
    assert div0.attrib == {"class": "row"}
    assert div0.text is None
    assert div0.tail is None
    assert len(div0) == 1

    div1 = tuple(div0)[0]
    assert isinstance(div1, ET.Element)
    assert div1.tag == "div"
    assert div1.attrib == {}
    assert div1.text is None
    assert div1.tail is None
    assert len(div1) == 1

    encoded_image = base64.b64encode(image).decode()

    img = tuple(div1)[0]
    assert isinstance(img, ET.Element)
    assert img.tag == "img"
    assert img.attrib == {"src": f"data:image/jpeg;base64,{encoded_image}"}
    assert img.text is None
    assert img.tail is None
    assert len(img) == 0


def test_obj_gif():
    image = svg_data.encode()
    class Repr:
        def _repr_mimebundle_(self) -> dict[str, Any]:
            return {"image/gif": image}

    content = rico._content.Obj(Repr(), class_="row")

    div0 = content.container
    assert isinstance(div0, ET.Element)
    assert div0.tag == "div"
    assert div0.attrib == {"class": "row"}
    assert div0.text is None
    assert div0.tail is None
    assert len(div0) == 1

    div1 = tuple(div0)[0]
    assert isinstance(div1, ET.Element)
    assert div1.tag == "div"
    assert div1.attrib == {}
    assert div1.text is None
    assert div1.tail is None
    assert len(div1) == 1

    encoded_image = base64.b64encode(image).decode()

    img = tuple(div1)[0]
    assert isinstance(img, ET.Element)
    assert img.tag == "img"
    assert img.attrib == {"src": f"data:image/gif;base64,{encoded_image}"}
    assert img.text is None
    assert img.tail is None
    assert len(img) == 0


def test_obj_text():
    class Repr:
        def _repr_mimebundle_(self) -> dict[str, Any]:
            return {"text/plain": "Hello world!"}

    content = rico._content.Obj(Repr(), class_="row")

    div0 = content.container
    assert isinstance(div0, ET.Element)
    assert div0.tag == "div"
    assert div0.attrib == {"class": "row"}
    assert div0.text is None
    assert div0.tail is None
    assert len(div0) == 1

    div1 = tuple(div0)[0]
    assert isinstance(div1, ET.Element)
    assert div1.tag == "div"
    assert div1.attrib == {}
    assert div1.text is None
    assert div1.tail is None
    assert len(div1) == 1

    p = tuple(div1)[0]
    assert isinstance(p, ET.Element)
    assert p.tag == "p"
    assert p.attrib == {}
    assert p.text == "Hello world!"
    assert p.tail is None
    assert len(p) == 0


def test_obj_priority():
    class Repr:
        def _repr_javascript_(self) -> str:
            return "alert('Hello World!');"

        def _repr_mimebundle_(self) -> dict[str, Any]:
            return {
                "text/plain": "Hello world!",
                "text/markdown": "# Hello world!",
            }

    content = rico._content.Obj(Repr(), class_="row")

    div0 = content.container
    assert isinstance(div0, ET.Element)
    assert div0.tag == "div"
    assert div0.attrib == {"class": "row"}
    assert div0.text is None
    assert div0.tail is None
    assert len(div0) == 1

    div1 = tuple(div0)[0]
    assert isinstance(div1, ET.Element)
    assert div1.tag == "div"
    assert div1.attrib == {}
    assert div1.text is None
    assert div1.tail is None
    assert len(div1) == 1

    h1 = tuple(div1)[0]
    assert isinstance(h1, ET.Element)
    assert h1.tag == "h1"
    assert h1.attrib == {}
    assert h1.text == "Hello world!"
    assert len(h1) == 0


def test_obj_corner_cases():
    class Repr:
        def _repr_javascript_(self) -> str | None:
            return None

        def _repr_markdown_(self) -> tuple[str, Any]:
            return "# Hello world!", "metadata"

    content = rico._content.Obj(Repr(), class_="row")

    div0 = content.container
    assert isinstance(div0, ET.Element)
    assert div0.tag == "div"
    assert div0.attrib == {"class": "row"}
    assert div0.text is None
    assert div0.tail is None
    assert len(div0) == 1

    div1 = tuple(div0)[0]
    assert isinstance(div1, ET.Element)
    assert div1.tag == "div"
    assert div1.attrib == {}
    assert div1.text is None
    assert div1.tail is None
    assert len(div1) == 1

    h1 = tuple(div1)[0]
    assert isinstance(h1, ET.Element)
    assert h1.tag == "h1"
    assert h1.attrib == {}
    assert h1.text == "Hello world!"
    assert len(h1) == 0



"""Container classes."""

from __future__ import annotations

import functools
from typing import TYPE_CHECKING
import xml.etree.ElementTree as ET

import rico._config
import rico._content
import rico._html


if TYPE_CHECKING:
    from collections.abc import Callable, Iterable
    from typing import Any, Concatenate, Literal, ParamSpec

    P = ParamSpec("P")


def _append(
    content_type: type[rico._content.ContentBase],
    init: Callable[Concatenate[Any, P], None],
) -> Callable[Concatenate[Any, P], None]:
    @functools.wraps(init)
    def method(
        self: Any,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> None:
        content = content_type(*args, **kwargs)
        self.container.append(content.container)

    return method

class Div(rico._content.Obj):
    def __init__(self, *objects: Any, class_: str | None = None):
        """Create a <div> container, create HTML elements and add them to the container.

        Each HTML element is also wrapped in a separate <div> container.

        Args:
            *objects: Objects that are used to create HTML elements.
            class_: The container class attribute.
        """
        super().__init__(*objects, class_=class_)

    append_tag = _append(rico._content.Tag, rico._content.Tag.__init__)
    append_text = _append(rico._content.Text, rico._content.Text.__init__)
    append_code = _append(rico._content.Code, rico._content.Code.__init__)
    append_html = _append(rico._content.HTML, rico._content.HTML.__init__)
    append_markdown = _append(rico._content.Markdown, rico._content.Markdown.__init__)
    append_image = _append(rico._content.Image, rico._content.Image.__init__)
    append_plot = _append(rico._content.Plot, rico._content.Plot.__init__)
    append_chart = _append(rico._content.Chart, rico._content.Chart.__init__)
    append = _append(rico._content.Obj, rico._content.Obj.__init__)


class Doc(Div):
    html: ET.Element
    head: ET.Element
    body: ET.Element

    def __init__(
        self,
        *objects: Any,
        title: str | None = None,
        bootstrap: Literal["css", "full", "none"] = "css",
        extra_styles: Iterable[rico._content.Style] = (),
        extra_scripts: Iterable[rico._content.Script] = (),
        class_: str | None = "container",
    ):
        """Create an HTML document.

        Creates a <div> container inside the <body> element.
        Create HTML elements and add them to the container.
        Each HTML element is also wrapped in a separate <div> container.

        Args:
            *objects: Objects that are used to create HTML elements.
            title: The document title.
            bootstrap: Defines which Bootatrap files are included in the document.
                If "css", include only CSS.
                If "full", include both the CSS and JS.
                If "none", don'e include Bootstrap.
            extra_styles: Extra styles to include in the document.
            extra_scripts: Extra scripts to include in the document.
            class_: The container class attribute.
        """
        super().__init__(*objects, class_=class_)
        self.html = ET.Element("html")
        self.head = ET.Element("head")
        self.body = ET.Element("body")
        self.html.append(self.head)
        self.html.append(self.body)
        self.body.append(self.container)

        if title is not None:
            title_element = ET.Element("title")
            title_element.text = title
            self.head.append(title_element)

        global_config = rico._config.get_config()
        if global_config["meta_charset"]:
            self.head.append(ET.Element("meta", charset=global_config["meta_charset"]))
        if global_config["meta_viewport"]:
            self.head.append(ET.Element(
                "meta",
                name="viewport",
                content=global_config["meta_viewport"],
            ))

        styles : list[rico._content.Style] = []
        scripts : list[rico._content.Script] = []

        if bootstrap.lower() in {"css", "full"}:
            styles.append(rico._content.Style(src=global_config["bootstrap_css"]))
        if bootstrap.lower() == "full":
            scripts.append(rico._content.Script(src=global_config["bootstrap_js"]))
        if global_config["dataframe_style"]:
            styles.append(rico._content.Style(text=global_config["dataframe_style"]))

        for style in (*styles, *extra_styles):
            self.head.append(style.container)

        for script in (*scripts, *extra_scripts):
            if script.footer:
                self.body.append(script.container)
            else:
                self.head.append(script.container)

    def serialize(
        self,
        indent: bool | None = None,
        space: str | None = None,
        strip: bool | None = None,
    ) -> str:
        """Serialize the HTML document to s string.

        Args:
            indent: If True, indent the elements.
            space: Whitespace for indentation.
            strip: If True, strip unnecessary whitespace.

        Returns:
            A string with serialized HTML.
        """
        return "<!doctype html>\n" + rico._html.serialize_html(
            self.html, indent=indent, space=space, strip=strip)
